'''
hep_spt
=======

Add extra functionality to the scikit-learn package,
allowing to keep track of the configurations used
to train MVA methods.
'''

__author__  = ['Miguel Ramos Pernas']
__email__   = ['miguel.ramos.pernas@cern.ch']


# Python
import importlib, inspect, os, pkgutil


__project_path__ = os.path.dirname(os.path.abspath(__file__))


__all__ = []
for loader, module_name, ispkg in pkgutil.walk_packages(__path__):

    __all__.append(module_name)

    if not ispkg:
        # Import all classes and functions
        mod = importlib.import_module('hep_spt.' + module_name)

        __all__ += mod.__all__

        for n, c in inspect.getmembers(mod):
            if n in mod.__all__:
                globals()[n] = c



'''
Function and classes representing statistical tools.
'''

__author__ = ['Miguel Ramos Pernas']
__email__  = ['miguel.ramos.pernas@cern.ch']


# Python
import os, warnings
import numpy as np
from math import exp, log, sqrt
from scipy.optimize import fsolve
from scipy.special import gamma
from scipy.stats import chi2, kstwobign
from scipy.stats import ks_2samp as scipy_ks_2samp

# Local
from hep_spt import __project_path__
from hep_spt.core import decorate

# Define confidence intervals.
__chi2_one_dof__ = chi2(1)
__one_sigma__    = __chi2_one_dof__.cdf(1)
__two_sigma__    = __chi2_one_dof__.cdf(4)
__three_sigma__  = __chi2_one_dof__.cdf(9)
__five_sigma__   = __chi2_one_dof__.cdf(25)

# Define Poisson tolerance. If this number is 100, we can
# use Stirling's approximation of log(k!) with a relative
# error of 0.001.
__poisson_from_stirling__ = 100

# Number after which the poisson uncertainty is considered to
# be the same as that of a gaussian with "std = sqrt(lambda)".
# This number has been chosen so the difference of the
# relative uncertainty is ~0.5%.
__poisson_to_gauss__ = 200


__all__ = ['jac_poisson_float_l', 'poisson_float', 'calc_poisson_freq_uncert',
           'ks_2samp', 'poisson_freq_uncert_one_sigma', 'process_uncert']


@decorate(np.vectorize)
def calc_poisson_freq_uncert( m, cl = __one_sigma__ ):
    '''
    Return the lower and upper bayesian uncertainties for
    a poisson distribution with mean "k" given the chi-square.

    :param m: mean of the Poisson distribution.
    :type m: float
    :param cl: confidence level (between 0 and 1).
    :type cl: float
    :returns: lower and upper uncertainties.
    :rtype: float, float
    '''
    sm = sqrt(m)

    if m > __poisson_to_gauss__:
        # Use the gaussian approximation of the uncertainty
        s = sqrt(m)
        return s, s

    alpha = (1. - cl)/2.

    if m < 1:
        # In this case there is only an upper uncertainty, so
        # the coverage is reset so it covers the whole "cl"
        lw = 0.
        alpha *= 2.
    else:
        ileft = m - sm
        if ileft < 0:
            ileft = 0.

        rleft = np.arange(m, m + 50*sm, dtype = int)
        fleft = lambda l: poisson_float(l, rleft).sum() - alpha
        jleft = lambda l: [jac_poisson_float_l(l, rleft).sum()]

        lw = fsolve(fleft, ileft, fprime = jleft)[0]

    iright = m + sm
    rright = np.arange(0, m + 1, dtype = int)
    fright = lambda l: poisson_float(l, rright).sum() - alpha
    jright = lambda l: [jac_poisson_float_l(l, rright).sum()]

    up = fsolve(fright, iright, fprime = jright)[0]

    return process_uncert(m, lw, up)


@decorate(np.vectorize)
def jac_poisson_float_l( l, k, tol = __poisson_from_stirling__ ):
    '''
    Return the value of Jacobian of the poisson_float
    function considering that it depends exclusively on "l".

    :param l: mean(s) of the Poisson.
    :type l: numpy.ndarray
    :param k: position(s) to evaluate.
    :type k: numpy.ndarray
    :param tol: tolerance(s) from which the Stirling's approximation \
    will be used.
    :type tol: numpy.ndarray
    :returns: value(s) of the poisson jacobian.
    :rtype: float
    '''
    if l == 0:
        return 0.

    return (k*1./l - 1.)*poisson_float(l, k, tol)


def _ks_2samp_values( arr, wgts = None ):
    '''
    Calculate the values needed to perform the Kolmogorov-Smirnov test.

    :param arr: input sample.
    :type arr: array-like
    :param wgts: possible weights.
    :type wgts: array-like
    :returns: sorted sample, stack with the cumulative distribution and
    sum of weights.
    :rtype: array-like, array-like, float
    '''
    wgts = wgts if wgts is not None else np.ones(len(arr), dtype=float)

    ix   = np.argsort(arr)
    arr  = arr[ix]
    wgts = wgts[ix]

    cs = np.cumsum(wgts)

    sw = cs[-1]

    hs = np.hstack((0, cs/sw))

    return arr, hs, sw


def ks_2samp( a, b, wa = None, wb = None ):
    '''
    Compute the Kolmogorov-Smirnov statistic on 2 samples. This is a two-sided
    test for the null hypothesis that 2 independent samples are drawn from the
    same continuous distribution. Weights for each sample are accepted. If no
    weights are provided, then the function scipy.stats.ks_2samp is called
    instead.

    :param a: first sample.
    :type a: array-like
    :param b: second sample.
    :type b: array-like
    :param wa: set of weights for "a". Same length as "a".
    :type wa: array-like or None.
    :param wb: set of weights for "b". Same length as "b".
    :type wb: array-like or None.
    :returns: test statistic and two-tailed p-value.
    :rtype: float, float
    '''
    if wa is None and wb is None:
        return scipy_ks_2samp(a, b)

    a, cwa, na = _ks_2samp_values(a, wa)
    b, cwb, nb = _ks_2samp_values(b, wb)

    m = np.concatenate([a, b])

    cdfa = cwa[np.searchsorted(a, m, side='right')]
    cdfb = cwb[np.searchsorted(b, m, side='right')]

    d = np.max(np.abs(cdfa - cdfb))

    en = np.sqrt(na*nb/float(na + nb))
    try:
        prob = kstwobign.sf((en + 0.12 + 0.11/en)*d)
    except:
        prob = 1.

    return d, prob


@decorate(np.vectorize)
def poisson_float( l, k, tol = __poisson_from_stirling__ ):
    '''
    Calculate the Poisson distribution value for floating
    numbers. The next term to the usual Stirling's
    approximation must be used to treat the cases where l ~ k.

    :param l: mean(s) of the Poisson distribution.
    :type l: numpy.ndarray
    :param k: position(s) to evaluate.
    :type k: numpy.ndarray
    :param tol: tolerance(s) from which the Stirling's approximation \
    will be used.
    :returns: value(s) of the Poisson distribution.
    :rtype: numpy.ndarray
    '''
    if l <= 0:
        return 0.

    if k > tol:
        return exp(k*log(l*1./k) + k - l - 0.5*log(2*np.pi*k))
    else:
        return 1./gamma(k + 1)*exp(k*log(l) - l)


def poisson_freq_uncert_one_sigma( m ):
    '''
    Return the poisson frequentist uncertainty at one standard
    deviation of confidence level. The input array recasted
    to int before doing the operation.

    :param m: measured value(s).
    :type m: array-like
    :returns: lower and upper frequentist uncertainties.
    :rtype: array-like(float, float)
    '''
    m = np.array(m, dtype = np.int32)

    out = np.zeros((len(m), 2), dtype = np.float64)

    ifile = os.path.join(__project_path__, 'data/poisson_freq_uncert_one_sigma.dat')

    table = np.loadtxt(ifile)

    no_app = (m < len(table))
    mk_app = np.logical_not(no_app)

    # Non-approximated uncertainties
    out[no_app] = table[m[no_app]]

    # Approximated uncertainties. This is not time-consuming. We
    # are just returning the square root of the input number.
    if mk_app.any():
        out[mk_app] = np.array(calc_poisson_freq_uncert(m[mk_app])).T

    return out


def process_uncert( m, lw, up ):
    '''
    Calculate the uncertainties and display an error if they
    have been incorrectly calculated.

    :param m: mean value.
    :type m: float
    :param lw: lower bound.
    :type lw: float
    :param up: upper bound.
    :type up: float
    :returns: lower and upper uncertainties.
    :type: array-like(float, float)
    '''
    s_lw = m - lw
    s_up = up - m

    if any(s < 0 for s in (s_lw, s_up)):
        warnings.warn('Poisson uncertainties have been '\
                      'incorrectly calculated')

    return s_lw, s_up


if __name__ == '__main__':
    '''
    Generate the tables to store the pre-calculated values of
    some uncertainties.
    '''
    conds = {
        'cl' : __one_sigma__,

        '__poisson_from_stirling__' : __poisson_from_stirling__,
        '__poisson_to_gauss__'      : __poisson_to_gauss__
        }

    m = np.arange(__poisson_to_gauss__)
    ucts = np.array(calc_poisson_freq_uncert(m, __one_sigma__)).T

    np.savetxt('data/poisson_freq_uncert_one_sigma.dat', ucts)



'''
Auxiliar mathematical functions.
'''

__author__ = ['Miguel Ramos Pernas']
__email__  = ['miguel.ramos.pernas@cern.ch']


# Custom
from hep_spt.core import decorate

# Python
import numpy as np
from functools import reduce


__all__ = ['gcd', 'is_power_2', 'lcm', 'next_power_2']


@decorate(np.vectorize)
def gcd( a, b, *args ):
    '''
    :param a: first number.
    :type a: int
    :param b: second number.
    :type b: int
    :param args: any other numbers.
    :type args: collection(int)
    :returns: greatest common divisor of "a" and "b".
    :rtype: int
    '''
    if len(args) == 0:
        while b:
            a, b = b, a % b
        return a
    else:
        return reduce(gcd, args + (a, b))


@decorate(np.vectorize)
def is_power_2( n ):
    '''
    :param n: input number.
    :type n: int
    :returns: whether the input number is a power of 2 or not.
    :rtype: bool
    '''
    return n > 0 and ((n & (n - 1)) == 0)


@decorate(np.vectorize)
def lcm( a, b, *args ):
    '''
    :param a: first number.
    :type a: int
    :param b: second number.
    :type b: int
    :param args: any other numbers.
    :type args: collection(int)
    :returns: least common multiple of "a" and "b".
    :rtype: int
    '''
    if len(args) == 0:
        return a*b//gcd(a, b)
    else:
        return reduce(lcm, args + (a, b))


@decorate(np.vectorize)
def next_power_2( n ):
    '''
    :param n: input number.
    :type n: int
    :returns: next power of 2 to the given number.
    :rtype: int

    .. note: If the input number is a power of two, it will return the \
    same number.
    '''
    return 1 << int(n - 1).bit_length()



'''
Some functions shared among modules.
'''

__author__ = ['Miguel Ramos Pernas']
__email__  = ['miguel.ramos.pernas@cern.ch']


# Python
from functools import wraps

__all__ = []


def decorate( deco ):
    '''
    Function to wrap a decorator so it preserves the name and
    docstring of the original function.

    :param deco: raw decorator which was meant to be used.
    :type deco: function-like
    :returns: decorator which preserves the name and docstring of the \
    original function.
    :rtype: function
    '''
    def _deco_wrapper( func ):
        '''
        This is the wrapper over the decorator.
        '''
        decorated_function = deco(func)

        @wraps(func)
        def _wrapper( *args, **kwargs ):
            '''
            Wrap the original function.
            '''
            return decorated_function(*args, **kwargs)

        return _wrapper

    return _deco_wrapper



#!/usr/bin/env python
'''
Setup script for the hep_spt package
'''

__author__ = 'Miguel Ramos Pernas'
__email__  = 'miguel.ramos.pernas@cern.ch'


# Python
import os
from setuptools import setup, find_packages

#
# Version of the package. Before a new release is made
# just the "version_info" must be changed. The options
# for the fourth tag are "dev", "alpha", "beta",
# "cand", "final" or "post".
#
version_info = (0, 0, 0, 'dev', 1)

tag = version_info[3]

if tag != 'final':
    if tag == 'alpha':
        frmt = '{}a{}'
    elif tag == 'beta':
        frmt = '{}b{}'
    elif tag == 'cand':
        frmt = '{}rc{}'
    elif tag == 'post':
        frmt = '{}.post{}'
    elif tag == 'dev':
        frmt = '{}.dev{}'
    else:
        raise ValueError('Unable to parse version information')

version = frmt.format('.'.join(map(str, version_info[:3])), version_info[4])

# Setup function
setup(

    name = 'hep_spt',

    version = version,

    description = 'Provides statistical and plotting tools using general '\
    'python packages, focused to High Energy Physics.',

    # Read the long description from the README
    long_description = open('README.rst').read(),

    # Keywords to search for the package
    keywords = 'physics hep statistics plotting',

    # Find all the packages in this directory
    packages = find_packages(),

    # Data files
    package_data = {'hep_spt': ['data/*', 'mpl/*']},

    # Requisites
    install_requires = ['matplotlib', 'numpy', 'pytest', 'scipy'],

    # Test requirements
    setup_requires = ['pytest-runner'],

    tests_require = ['pytest'],
    )


# Create a module with the versions
version_file = open('hep_spt/version.py', 'wt')
version_file.write("""\
'''
Auto-generated module holding the version of the hep_spt package
'''

__version__ = "{}"
__version_info__ = {}

__all__ = ['__version__', '__version_info__']
""".format(version, version_info))
version_file.close()



'''
Provide some useful functions to plot with matplotlib.
'''

__author__ = ['Miguel Ramos Pernas']
__email__  = ['miguel.ramos.pernas@cern.ch']


# Local
from hep_spt import __project_path__
from hep_spt.math_aux import lcm
from hep_spt.stats import poisson_freq_uncert_one_sigma

# Python
import matplotlib.pyplot as plt
import numpy as np
import math, os
from cycler import cycler


__all__ = [
    'PlotVar',
    'available_styles',
    'errorbar_hist',
    'opt_fig_div',
    'path_to_styles',
    'process_range',
    'pull',
    'samples_cycler',
    'set_style',
    'text_in_rectangles'
    ]

# Path to the directory containing the styles
__path_to_styles__ = os.path.join(__project_path__, 'mpl')


class PlotVar:
    '''
    Define an object to store plotting information, like a name, a title,
    a binning scheme and a range.
    '''
    def __init__( self, name, title = None, bins = 20, rg = None, logscale = False ):
        '''
        :param name: name of the variable.
        :type name: str
        :param title: title of the variable.
        :type title: str
        :param bins: see the argument "bins" in :func:`numpy.histogram`.
        :type bins: int or sequence of scalars or str
        :param rg: range of this variable (min, max).
        :type rg: tuple(float, float)
        :param logscale: store whether to use logscale in the plots.
        :type logscale: bool
        '''
        self.name     = name
        self.title    = title or name
        self.bins     = bins
        self.rg       = rg
        self.logscale = logscale


def available_styles():
    '''
    :returns: list with the names of the available styles within this package.
    :rtype: list(str)
    '''
    available_styles = list(map(lambda s: s[:s.find('.mplstyle')],
                                os.listdir(__path_to_styles__)))
    return available_styles


def errorbar_hist( arr, bins = 20, rg = None, wgts = None, norm = False ):
    '''
    Calculate the values needed to create an error bar histogram.

    :param arr: input array of data to process.
    :param bins: see :func:`numpy.histogram`.
    :type bins: int or sequence of scalars or str
    :param rg: range to process in the input array.
    :type rg: tuple(float, float)
    :param wgts: possible weights for the histogram.
    :type wgts: collection(value-type)
    :param norm: if True, normalize the histogram. If it is set to a number, \
    the histogram is normalized and multiplied by that number.
    :type norm: bool, int or float
    :returns: values, edges, the spacing between bins in X the Y errors. \
    In the non-weighted case, errors in Y are returned as two arrays, with the \
    lower and upper uncertainties.
    :rtype: numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray
    '''
    if wgts is not None:
        # Use sum of the square of weights to calculate the error
        sw2, edges = np.histogram(arr, bins, rg, weights = wgts*wgts)

        values, _ = np.histogram(arr, edges, weights = wgts)

        ey = np.sqrt(sw2)

    else:
        # Use the poissonian errors
        values, edges = np.histogram(arr, bins, rg, weights = wgts)

        # For compatibility with matplotlib.pyplot.errorbar
        ey = poisson_freq_uncert_one_sigma(values).T

    ex = (edges[1:] - edges[:-1])/2.

    if norm:

        s = float(values.sum())/norm

        if s != 0:
            values = values/s
            ey = ey/s
        else:
            ey *= np.finfo(ey.dtype).max

    return values, edges, ex, ey


def opt_fig_div( naxes ):
    '''
    Get the optimal figure division for a given number of axes, where
    all the axes have the same dimensions.

    :param naxes: number of axes to plot in the figure.
    :type naxes: int
    :returns: number of rows and columns of axes to draw.
    :rtype: int, int
    '''
    nstsq = int(round(math.sqrt(naxes)))

    if nstsq**2 > naxes:
        nx = nstsq
        ny = nstsq
    else:
        nx = nstsq
        ny = nstsq
        while nx*ny < naxes:
            ny += 1

    return nx, ny


def path_to_styles():
    '''
    :returns: path to the directory containing the styles.
    :rtype: str
    '''
    return __path_to_styles__


def process_range( arr, rg = None ):
    '''
    Process the given range, determining the minimum and maximum
    values for a 1D histogram.

    :param arr: array of data.
    :type arr: numpy.ndarray
    :param rg: range of the histogram. It must contain tuple(min, max), \
    where "min" and "max" can be either floats (1D case) or collections \
    (ND case).
    :type rg: tuple or None
    :returns: minimum and maximum values.
    :rtype: float, float
    '''
    if rg is not None:
        vmin, vmax = rg
    else:
        amax = arr.max(axis = 0)
        vmin = arr.min(axis = 0)
        vmax = np.nextafter(amax, np.infty)

    return vmin, vmax


def pull( vals, err, ref ):
    '''
    Get the pull with the associated errors for a given set of values and a
    reference. Considering, :math:`v` as the experimental value and :math:`r`
    as the rerference, the definition of this quantity is :math:`(v - r)/\sigma`
    in case symmetric errors are provided. In the case of asymmetric errors the
    definition is:

    .. math::
       \\text{pull}
       =
       \Biggl \lbrace
       {
       (v - r)/\sigma_{low},\\text{ if } v - r \geq 0
       \\atop
       (v - r)/\sigma_{up}\\text{ otherwise }
       }

    In the latter case, the errors are computed in such a way that the closer to
    the reference is equal to 1 and the other is scaled accordingly, so if
    :math:`v - r > 0`, then :math:`\sigma^{pull}_{low} = 1` and
    :math:`\sigma^{pull}_{up} = \sigma_{up}/\sigma_{low}`.

    :param vals: values to compare with.
    :type vals: array-like
    :param err: array of errors. Both symmetric and asymmetric errors \
    can be provided. In the latter case, they must be provided as a \
    (2, n) array.
    :type err: array-like
    :param ref: reference to follow.
    :type ref: array-like
    :returns: pull of the values with respect to the reference and \
    associated errors. In case asymmetric errors have been provided, \
    the returning array has shape (2, n).
    :rtype: array-like, array-like
    '''
    pull = vals - ref

    perr = np.ones_like(err)

    if len(err.shape) == 1:
        # Symmetric errors
        pull /= err

    elif len(err.shape) == 2:
        # Asymmetric errors

        up = (pull >= 0)
        lw = (pull < 0)

        el, eu = err

        pull[up] /= el[up]
        pull[lw] /= eu[lw]

        perr_l, perr_u = perr

        perr_l[lw] = (el[lw]/eu[lw])
        perr_u[up] = (eu[up]/el[up])
    else:
        raise TypeError('The error array must have shape (2, n) or (n,)')

    return pull, perr


def samples_cycler( smps, *args, **kwargs ):
    '''
    Often, one wants to plot several samples with different matplotlib
    styles. This function allows to create a cycler.cycler object
    to loop over the given samples, where the "label" key is filled
    with the values from "smps".

    :param smps: list of names for the samples.
    :type smps: collection(str)
    :param args: position argument to cycler.cycler.
    :type args: tuple
    :param kwargs: keyword arguments to cycler.cycler.
    :type kwargs: dict
    :returns: cycler object with the styles for each sample.
    :rtype: cycler.cycler
    '''
    cyc = cycler(*args, **kwargs)

    ns = len(smps)
    nc = len(cyc)

    if ns > nc:

        warnings.warn('Not enough plotting styles in cycler, '\
                      'some samples might have the same style.')

        l = math_aux.lcm(ns, nc)

        re_cyc = (l*cyc)[:ns]
    else:
        re_cyc = cyc[:ns]

    return re_cyc + cycler(label = smps)


def set_style( *args ):
    '''
    Set the style for matplotlib to one within this project. Available styles
    are:

    * singleplot: designed to create a single figure.
    * multiplot: to make subplots. Labels and titles are smaller than in \
    "singleplot", although lines and markers maintain their sizes.

    By default the "singleplot" style is set.

    :param args: styles to load.
    :type args: tuple
    '''
    args = list(args)
    if len(args) == 0:
        # The default style is always set
        args = ['default', 'singleplot']
    elif 'default' not in args:
        args.insert(0, 'default')

    avsty = available_styles()

    sty_args = []
    for s in args:
        if s not in avsty:
            warnings.warn('Unknown style "{}", will not be loaded'.format(style))
        else:
            sty_args.append(os.path.join(__path_to_styles__, '{}.mplstyle'.format(s)))

    plt.style.use(sty_args)


def text_in_rectangles( ax, recs, txt, **kwargs ):
    '''
    Write text inside matplotlib.patches.Rectangle instances.

    :param ax: axes where the rectangles are being drawn.
    :type ax: matplotlib.axes.Axes
    :param recs: set of rectangles to work with.
    :type recs: collection(matplotlib.patches.Rectangle)
    :param txt: text to fill with in each rectangle.
    :type txt: collection(str)
    :param kwargs: any other argument to matplotlib.axes.Axes.annotate.
    :type kwargs: dict
    '''
    for r, t in zip(recs, txt):
        x, y = r.get_xy()

        cx = x + r.get_width()/2.
        cy = y + r.get_height()/2.

        ax.annotate(t, (cx, cy), **kwargs)



'''
Module to manage special histograms, like adaptive binned 1D and 2D histograms.
'''

__author__ = ['Miguel Ramos Pernas']
__email__  = ['miguel.ramos.pernas@cern.ch']


# Custom
from hep_spt.plotting import errorbar_hist, process_range
from hep_spt.stats import poisson_freq_uncert_one_sigma

# Python
import numpy as np
import bisect, itertools
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle


__all__ = ['AdBin', 'adbin_as_rectangle', 'adbin_hist2d_rectangles',
           'adbin_hist1d', 'adbin_hist1d_edges',
           'adbin_hist2d', 'adbin_histnd'
           ]


class AdBin:
    '''
    Represent a n-dimensional adaptive bin. This class is meant so serve
    as interface between the user and matplotlib to plot adaptive
    binned histograms.
    '''
    def __init__( self, arr, rg = None, wgts = None ):
        '''
        :param arr: array of data.
        :type arr: numpy.ndarray
        :param rg: range of the histogram in each dimension. As \
        [(xmin, ymin), (xmax, ymax)].
        :type rg: numpy.ndarray(float, float) or None.
        :param wgts: possible weights.
        :type wgts: numpy.ndarray or None
        '''
        arr, (vmin, vmax), wgts = _proc_hist_input(arr, rg, wgts)

        self.arr  = arr
        self.wgts = wgts
        self.vmin = np.array(vmin, dtype = float)
        self.vmax = np.array(vmax, dtype = float)

    def contains( self, arr ):
        '''
        :param arr: input data.
        :type arr: numpy.ndarray
        :returns: whether the values in the input array are inside this bin or \
        not.
        :rtype: bool or numpy.ndarray(bool)
        '''
        return np.logical_and(arr >= self.vmin, arr < self.vmax).all(axis = 1)

    def dens( self, arr, wgts = None ):
        '''
        :param arr: array of data to process.
        :type arr: numpy.ndarray
        :param wgts: possible weights.
        :type wgts: numpy.ndarray or None
        :returns: density of this bin.
        :rtype: float
        '''
        return self.sw(arr, wgts)/float(self.size())

    def divide( self, ndiv = 2 ):
        '''
        Divide this bin in two, using the median in each dimension. The
        dimension used to make the division is taken as that which generates
        the smallest bin.

        :param ndiv: number of divisions to create. For large values, this \
        algorithm will ask for having a low sum of weights for the first \
        bin, which will translate in having a long thin bin.
        :type ndiv: int
        :returns: two new bins, supposed to contain half the sum of weights of \
        the parent.
        :rtype: AdBin, AdBin
        '''
        assert ndiv > 1

        srt   = self.arr.argsort(axis = 0)
        sarr  = np.array([self.arr.T[i][s] for i, s in enumerate(srt.T)]).T
        swgts = self.wgts[srt]
        csw   = swgts.cumsum(axis = 0)

        co = np.max(csw)/float(ndiv)

        p = np.array([bisect.bisect_left(c, co) for c in csw.T])

        bounds = np.array([np.nextafter(sarr[p,i], np.infty)
                           for i, p in enumerate(p)])

        mask_left  = (self.arr < bounds)
        mask_right = (self.arr >= bounds)

        # These functions calculate the sizes of the bins generated
        # with each cut.
        size = lambda arr: arr.max() - arr.min()

        def _msz( arr, i ):
            '''
            Calculate the minimum size of the input array, for the
            given index and considering the two global masks
            "mask_left" and "mask_right".
            '''
            sarr = arr[:,i]
            sz   = float(size(sarr))

            sl = size(sarr[mask_left[:,i]])/sz
            sr = size(sarr[mask_right[:,i]])/sz

            return min(sl, sr)

        frags = np.array([_msz(self.arr, i) for i in range(self.arr.shape[1])])

        # The sample is cut following the criteria that leads to the
        # smallest bin possible.

        min_dim = frags.argmin()

        il = mask_left[:,min_dim]
        ir = mask_right[:,min_dim]

        left   = self.arr[il]
        right  = self.arr[ir]
        wleft  = self.wgts[il]
        wright = self.wgts[ir]

        d = bounds[min_dim]

        lbd = np.array(self.vmax)
        lbd[min_dim] = d
        bl = AdBin(left, (self.vmin, lbd), wleft)

        rbd = np.array(self.vmin)
        rbd[min_dim] = d
        br = AdBin(right, (rbd, self.vmax), wright)

        # If the number of divisions is greater than 2, perform again the same
        # operation in the bin on the right
        all_bins = [bl]
        if ndiv > 2:
            all_bins += br.divide(ndiv - 1)
        else:
            all_bins.append(br)

        return all_bins

    def size( self ):
        '''
        :returns: size of this bin calculated as the product of \
        the individual sizes in each dimension.
        :rtype: float
        '''
        return float(np.prod(self.vmax - self.vmin))

    def sw( self, arr, wgts = None ):
        '''
        :param arr: array of data to process.
        :type arr: numpy.ndarray
        :param wgts: possible weights.
        :type wgts: numpy.ndarray or None
        :returns: sum of weights for this bin.
        :rtype: float
        '''
        true = self.contains(arr)

        if wgts is not None:
            sw = wgts[true].sum()
        else:
            sw = np.count_nonzero(true)

        return float(sw)

    def sw_uncert( self, arr, wgts = None ):
        '''
        :param arr: array of data to process.
        :type arr: numpy.ndarray
        :param wgts: possible weights.
        :type wgts: numpy.ndarray or None
        :returns: uncertainty of the sum of weights in the bin. If "wgts" is \
        provided, this magnitude is equal to the square root of the sum of \
        weights in the bin. Otherwise poissonian errors are considered.
        :rtype: float
        '''
        true = self.contains(arr)

        if wgts is not None:
            uncert = np.sqrt((wgts*wgts)[true].sum())
        else:
            sw = np.count_nonzero(true)

            uncert = poisson_freq_uncert_one_sigma(sw)

        return float(uncert)


def adbin_as_rectangle( adb, **kwargs ):
    '''
    Extract the bounds so it can be processed by a matplotlib.patches.Rectangle
    object and properly fill the area inside it.

    :param adb: input adaptive bin.
    :type adb: AdBin
    :param kwargs: extra arguments to matplotlib.patches.Rectangle.
    :type kwargs: dict
    :returns: rectangle to be drawn with matplotlib.
    :rtype: matplotlib.patches.Rectangle
    '''
    xmin, ymin = adb.vmin
    xmax, ymax = adb.vmax

    width  = xmax - xmin
    height = ymax - ymin

    return Rectangle((xmin, ymin), width, height, **kwargs)


def adbin_hist1d( arr, nbins = 100, rg = None, wgts = None, **kwargs ):
    '''
    Create an adaptive binned histogram.

    :param arr: array of data.
    :type arr: numpy.ndarray
    :param nbins: number of bins.
    :type nbins: int
    :param rg: range of the histogram.
    :type rg: tuple(float, float) or None
    :param wgts: optional array of weights.
    :type wgts: numpy.ndarray or None
    :param kwargs: any other argument to :func:`plotting.errorbar_hist`.
    :type kwargs: dict
    :returns: values, edges, the spacing between bins in X the Y errors. \
    In the non-weighted case, errors in Y are returned as two arrays, with the \
    lower and upper uncertainties.
    :rtype: numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray
    '''
    arr, rg, pws = _proc_hist_input_1d(arr, rg, wgts)

    # Sort the data
    srt = arr.argsort()
    arr = arr[srt]
    pws = pws[srt]

    # Solving the problem from the left and from the right reduces
    # the bias in the last edges
    le = adbin_hist1d_edges(arr, nbins, rg, pws)
    re = adbin_hist1d_edges(arr[::-1], nbins, rg, pws[::-1])[::-1]

    edges = (re + le)/2.

    vmin, vmax = rg
    edges[0]   = vmin
    edges[-1]  = vmax

    return errorbar_hist(arr, edges, rg, wgts, **kwargs)


def adbin_hist1d_edges( arr, nbins = 100, rg = None, wgts = None ):
    '''
    Create adaptive binned edges from the given array.

    :param arr: array of data.
    :type arr: numpy.ndarray
    :param nbins: number of bins.
    :type nbins: int
    :param rg: range of the histogram.
    :type rg: tuple(float, float) or None
    :param wgts: optional array of weights.
    :type wgts: numpy.ndarray or None
    :returns: edges of the histogram, with size (nbins + 1).
    :rtype: numpy.ndarray
    '''
    arr, (vmin, vmax), wgts = _proc_hist_input_1d(arr, rg, wgts)

    edges = np.zeros(nbins + 1)

    for i in range(nbins - 1):

        csum = wgts.cumsum()
        reqs = csum[-1]/float(nbins - i)

        p = bisect.bisect_left(csum, reqs)
        if p == len(csum):
            p = -1

        s = csum[p]

        if s != reqs:
            # If the sum differs, then decide whether to use the
            # current index based on the relative difference.
            if p != 0 and np.random.uniform() < (s - reqs)/reqs:
                p -= 1

        edges[i + 1]  = (arr[p] + arr[p + 1])/2.

        arr  = arr[p + 1:]
        wgts = wgts[p + 1:]

    edges[0]  = vmin
    edges[-1] = vmax

    return edges


def adbin_hist2d( x, y, *args, **kwargs ):
    '''
    Create a 2D adaptive binned histogram. This function calls
    the adbin_histnd function.

    :param arr: array of data with the variables as columns.
    :type arr: numpy.ndarray
    :param nbins: number of bins.
    :type nbins: int
    :param wgts: optional array of weights.
    :type wgts: numpy.ndarray
    :returns: adaptive bins of the histogram, with size (nbins + 1).
    :rtype: list(AdBin)

    .. seealso: :func:`adbin_hist2d`
    '''
    return adbin_histnd(np.array([x, y]).T, *args, **kwargs)


def adbin_hist2d_rectangles( bins, arr,
                             rg = None, wgts = None,
                             cmap = None, fill = 'sw',
                             color = True,
                             **kwargs ):
    '''
    Create a collection of rectangles from a collection of bins, using
    the input data to calculate the associated quantity (specified in "fill").

    :param bins: input bins.
    :type bins: collection(AdBin)
    :param arr: input data.
    :type arr: collection(value-type)
    :param rg: range of the histogram in each dimension. As \
    [(xmin, ymin), (xmax, ymax)].
    :type rg: tuple(np.ndarray, np.ndarray) or None
    :param wgts: input weights.
    :type wgts: collection(value-type)
    :param cmap: optional color map. If "None", the default from \
    matplotlib.pyplot is used.
    :type cmap: matplotlib.colors.Colormap or None
    :param fill: method to use for filling ('sw' or 'dens').
    :type fill: str
    :param color: whether the output rectangles are filled with a color or not.
    :type color: bool
    :param kwargs: any other argument to :func:`adbin_as_rectangle`.
    :type kwargs: dict
    :returns: rectangles and contents.
    :rtype: numpy.ndarray, numpy.ndarray
    '''
    assert fill in ('sw', 'dens')

    recs = [adbin_as_rectangle(b, **kwargs) for b in bins]

    arr, rg, wgts = _proc_hist_input(arr, rg, wgts)

    # Get the contents associated to each bin
    func = getattr(AdBin, fill)
    contents = np.array(list(map(lambda b: func(b, arr, wgts), bins)))

    if color:

        cmap = cmap or plt.get_cmap()

        vmin = contents.min()
        vmax = contents.max()

        sz = float(vmax - vmin)

        # We need to normalize the color map. The normalization
        # depends if the input number is an integer (0, length of
        # the color map) or a float (0., 1.).
        for r, c in zip(recs, contents):
            if sz != 0:
                r.set_facecolor(cmap((c - vmin)/sz))
            else:
                r.set_facecolor('none')
    else:
        for r in recs:
            r.set_facecolor((1, 1, 1))

    return recs, contents


def adbin_histnd( arr, nbins = 100, rg = None, wgts = None, ndiv = 2 ):
    '''
    Create a ND adaptive binned histogram.

    :param arr: array of data with the variables as columns.
    :type arr: numpy.ndarray
    :param rg: range of the histogram in each dimension. As \
    [(xmin, ymin), (xmax, ymax)].
    :type rg: numpy.ndarray(float, float) or None.
    :param nbins: number of bins. In this algorithm, divisions will be made \
    till the real number of bins is equal or greater than "nbins". If this \
    number is a power of "ndiv", then the real number of bins will match \
    "nbins".
    :type nbins: int
    :param wgts: optional array of weights.
    :type wgts: numpy.ndarray or None
    :param ndiv: see :meth:`AdBin.divide`.
    :type ndiv: int
    :returns: adaptive bins of the histogram, with size (nbins + 1).
    :rtype: list(AdBin)

    .. seealso: :meth:`AdBin.divide`
    '''
    assert len(arr) // nbins > 0

    bins = [AdBin(arr, rg, wgts)]
    while len(bins) < nbins:
        bins = list(itertools.chain.from_iterable(a.divide(ndiv) for a in bins))

    return bins


def _proc_hist_input_1d( arr, rg = None, wgts = None ):
    '''
    Process some of the input arguments of the functions to
    manage 1D histograms.

    :param arr: array of data.
    :type arr: numpy.ndarray
    :param rg: range of the histogram.
    :type rg: tuple(float, float) or None
    :param wgts: optional array of weights.
    :type wgts: numpy.ndarray or None
    :returns: processed array of data, weights, and the minimum \
    and maximum values.
    :rtype: numpy.ndarray, tuple(float, float), numpy.ndarray
    '''
    vmin, vmax = process_range(arr, rg)

    cond = np.logical_and(arr >= vmin, arr < vmax)

    arr = arr[cond]

    if wgts is not None:
        wgts = wgts[cond]
    else:
        wgts = np.ones(len(arr))

    return arr, (vmin, vmax), wgts


def _proc_hist_input( arr, rg = None, wgts = None ):
    '''
    Process some of the input arguments of the functions to
    manage ND histograms.

    :param arr: array of data.
    :type arr: numpy.ndarray
    :param rg: range of the histogram in each dimension. As \
    [(xmin, ymin), (xmax, ymax)].
    :type rg: tuple(np.ndarray, np.ndarray) or None
    :param wgts: optional array of weights.
    :type wgts: numpy.ndarray or None
    :returns: processed array of data, weights, and the minimum \
    and maximum values for each dimension.
    :rtype: numpy.ndarray, tuple(np.ndarray, np.ndarray), numpy.ndarray
    '''
    vmin, vmax = process_range(arr, rg)

    cond = np.logical_and(arr >= vmin, arr < vmax).all(axis = 1)

    arr = arr[cond]

    if wgts is not None:
        wgts = wgts[cond]
    else:
        wgts = np.ones(len(arr))

    return arr, (vmin, vmax), wgts
